# --------------------------------------------------------------------------- #
#   list.py                                                                   #
#                                                                             #
#   Copyright © 2015-2020, Rajiv Bakulesh Shah, original author.              #
#   All rights reserved.                                                      #
# --------------------------------------------------------------------------- #


import collections.abc
import functools
import itertools

from redis import ResponseError

from .base import Base
from .exceptions import KeyExistsError


class RedisList(Base, collections.abc.MutableSequence):
    'Redis-backed container compatible with Python lists.'

    def _raise_on_error(func):
        @functools.wraps(func)
        def wrap(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except ResponseError:
                raise IndexError('list assignment index out of range')
        return wrap

    def _slice_to_indices(self, slice_or_index):
        try:
            start = slice_or_index.start or 0
            stop = slice_or_index.stop or len(self)
            step = slice_or_index.step or 1
        except AttributeError:
            start = slice_or_index
            stop = slice_or_index + 1
            step = 1
        indices = range(start, stop, step)
        return indices

    def __init__(self, iterable=tuple(), *, redis=None, key=None):
        'Initialize a RedisList.  O(n)'
        super().__init__(iterable, redis=redis, key=key)
        with self._watch(iterable):
            self._populate(iterable)

    def _populate(self, iterable=tuple()):
        encoded_values = [self._encode(value) for value in iterable]
        if encoded_values:
            if self.redis.exists(self.key):
                raise KeyExistsError(self.redis, self.key)
            else:
                self.redis.multi()
                self.redis.rpush(self.key, *encoded_values)

    # Methods required by collections.abc.MutableSequence:

    def __getitem__(self, index):
        'l.__getitem__(index) <==> l[index].  O(n)'
        with self._watch():
            if isinstance(index, slice):
                # This is monumentally stupid.  Python's list API requires us
                # to get elements by slice (defined as a start index, a stop
                # index, and a step).  Of course, Redis allows us to only get
                # elements by start and stop (no step), because it's Redis.  So
                # our ridiculous hack is to get all of the elements between
                # start and stop from Redis, then discard the ones between step
                # in Python.  More info:
                # http://redis.io/commands/lrange
                indices = self._slice_to_indices(index)
                self.redis.multi()
                self.redis.lrange(self.key, indices[0], indices[-1])
                encoded = self.redis.execute()[0]
                encoded = encoded[::index.step]
                value = [self._decode(value) for value in encoded]
            else:
                self.redis.multi()
                self.redis.lindex(self.key, index)
                encoded = self.redis.execute()[0]
                if encoded is None:
                    raise IndexError('list index out of range')
                else:
                    value = self._decode(encoded)
            return value

    @_raise_on_error
    def __setitem__(self, index, value):
        'l.__setitem__(index, value) <==> l[index] = value.  O(n)'
        with self._watch():
            if isinstance(index, slice):
                encoded_values = [self._encode(value) for value in value]
                indices = self._slice_to_indices(index)
                self.redis.multi()
                for index, encoded_value in zip(indices, encoded_values):
                    self.redis.lset(self.key, index, encoded_value)
                indices, num = indices[len(encoded_values):], 0
                for index in indices:
                    self.redis.lset(self.key, index, 0)
                    num += 1
                if num:
                    self.redis.lrem(self.key, num, 0)
            else:
                self.redis.multi()
                self.redis.lset(self.key, index, self._encode(value))

    @_raise_on_error
    def __delitem__(self, index):
        'l.__delitem__(index) <==> del l[index].  O(n)'
        # This is monumentally stupid.  Python's list API requires us to
        # delete an element by *index.*  Of course, Redis doesn't support
        # that, because it's Redis.  Instead, Redis supports deleting an
        # element by *value.*  So our ridiculous hack is to set l[index] to
        # None, then to delete the value None.  More info:
        # http://redis.io/commands/lrem
        with self._watch():
            self._delete(index)

    def _delete(self, index):
        indices, num = self._slice_to_indices(index), 0
        self.redis.multi()
        for index in indices:
            self.redis.lset(self.key, index, 0)
            num += 1
        if num:  # pragma: no cover
            self.redis.lrem(self.key, num, 0)

    def __len__(self):
        'Return the number of items in a RedisList.  O(1)'
        return self.redis.llen(self.key)

    def insert(self, index, value):
        'Insert an element into a RedisList before the given index.  O(n)'
        with self._watch():
            self._insert(index, value)

    def _insert(self, index, value):
        encoded_value = self._encode(value)
        if index <= 0:
            self.redis.multi()
            self.redis.lpush(self.key, encoded_value)
        elif index < len(self):
            # This is monumentally stupid.  Python's list API requires us
            # to insert an element before the given *index.*  Of course,
            # Redis doesn't support that, because it's Redis.  Instead,
            # Redis supports inserting an element before a given (pivot)
            # *value.*  So our ridiculous hack is to set the pivot value to
            # None, then to insert the desired value and the original pivot
            # value before the value None, then to delete the value None.
            # More info:
            # http://redis.io/commands/linsert
            pivot = self._encode(self[index])
            self.redis.multi()
            self.redis.lset(self.key, index, 0)
            for encoded_value in (encoded_value, pivot):
                self.redis.linsert(self.key, 'BEFORE', 0, encoded_value)
            self.redis.lrem(self.key, 1, 0)
        else:
            self.redis.multi()
            self.redis.rpush(self.key, encoded_value)

    # Methods required for Raj's sanity:

    def sort(self, *, key=None, reverse=False):
        'Sort a RedisList in place.  O(n)'
        if key is None:
            self.redis.sort(self.key, desc=reverse, store=self.key)
        else:
            raise NotImplementedError('sorting by key not implemented')

    def __eq__(self, other):
        if super().__eq__(other):
            # self and other are both RedisLists on the same Redis instance and
            # with the same key.  No need to compare element by element.
            return True
        else:
            with self._watch(other):
                try:
                    if len(self) != len(other):
                        # self and other are different lengths.
                        return False
                    elif isinstance(other, collections.abc.Sequence):
                        # self and other are the same length, and other is an
                        # ordered collection too.  Compare self's and other's
                        # elements, pair by pair.
                        for value1, value2 in zip(self, other):
                            if value1 != value2:
                                return False
                        else:
                            return True
                    else:
                        # self and other are the same length, but other is an
                        # unordered collection.
                        return False
                except TypeError:
                    return False

    def __add__(self, other):
        'Append the items in other to a RedisList.  O(n)'
        with self._watch(other):
            iterable = itertools.chain(self, other)
            return self.__class__(iterable, redis=self.redis)

    def __repr__(self):
        'Return the string representation of a RedisList.  O(n)'
        encoded = self.redis.lrange(self.key, 0, -1)
        values = [self._decode(value) for value in encoded]
        return self.__class__.__name__ + str(values)

    # Method overrides:

    # From collections.abc.MutableSequence:
    def append(self, value):
        'Add an element to the right side of the RedisList.  O(1)'
        self.extend((value,))

    # From collections.abc.MutableSequence:
    def extend(self, values):
        'Extend a RedisList by appending elements from the iterable.  O(1)'
        with self._watch(values):
            encoded_values = (self._encode(value) for value in values)
            self.redis.multi()
            self.redis.rpush(self.key, *encoded_values)

    # From collections.abc.MutableSequence:
    def pop(self, index=None):
        with self._watch():
            len_ = len(self)
            if index and index >= len_:
                raise IndexError('pop index out of range')
            elif index in {0, None, len_-1, -1}:
                pop_method = 'lpop' if index == 0 else 'rpop'
                self.redis.multi()
                getattr(self.redis, pop_method)(self.key)
                encoded_value = self.redis.execute()[0]
                if encoded_value is None:
                    raise IndexError(
                        'pop from an empty {}'.format(self.__class__.__name__),
                    )
                else:
                    return self._decode(encoded_value)
            else:
                value = self[index]
                self._delete(index)
                return value

    # From collections.abc.MutableSequence:
    def remove(self, value):
        with self._watch():
            for index, element in enumerate(self):
                if element == value:
                    self._delete(index)
                    break
            else:
                raise ValueError(
                    '{class_}.remove(x): x not in {class_}'.format(
                        class_=self.__class__.__name__,
                    ),
                )



# --------------------------------------------------------------------------- #
#   monkey.py                                                                 #
#                                                                             #
#   Copyright © 2015-2020, Rajiv Bakulesh Shah, original author.              #
#   All rights reserved.                                                      #
# --------------------------------------------------------------------------- #
'Monkey patches.'


import logging


_logger = logging.getLogger('pottery')


# Monkey patch os.listdir() to optionally return absolute paths.

import os

def _absolutize(*, files, path=None):
    path = os.path.abspath(path or '.')
    files = [os.path.join(path, f) for f in files]
    return files

def _listdir(path=None, *, absolute=False):
    files = _listdir.listdir(path)
    if absolute:  # pragma: no cover
        files = _absolutize(path=path, files=files)
    return files

_listdir.listdir = os.listdir
os.listdir = _listdir

_logger.info('Monkey patched os.listdir() to optionally return absolute paths')


# The Redis client doesn't have a sane equality test.  So monkey patch equality
# comparisons on to the Redis client.  We consider two Redis clients to be
# equal if they're connected to the same host, port, and database.
#
# I've submitted this change to redis-py:
#     https://github.com/andymccurdy/redis-py/pull/1240
#
# If it gets merged upstream, then I'll be able to delete this monkey patch.

from redis import ConnectionPool
from redis import Redis

def __eq__(self, other):
    try:
        return self.connection_kwargs == other.connection_kwargs
    except AttributeError:  # pragma: no cover
        return False

ConnectionPool.__eq__ = __eq__

def __eq__(self, other):
    '''True if two Redis clients are equal.

    The Redis client doesn't have a sane equality test.  So we monkey patch
    this method on to the Redis client so that two client instances are equal
    if they're connected to the same Redis host, port, and database.
    '''
    try:
        return self.connection_pool == other.connection_pool
    except AttributeError:
        return False

Redis.__eq__ = __eq__

_logger.info(
    'Monkey patched ConnectionPool.__eq__() and Redis.__eq__() to compare '
    'Redis clients by connection params'
)



# --------------------------------------------------------------------------- #
#   cache.py                                                                  #
#                                                                             #
#   Copyright © 2015-2020, Rajiv Bakulesh Shah, original author.              #
#   All rights reserved.                                                      #
# --------------------------------------------------------------------------- #


import collections
import functools
import logging

from redis import Redis
from redis.exceptions import WatchError

from .base import random_key
from .dict import RedisDict


_DEFAULT_TIMEOUT = 60   # seconds

_logger = logging.getLogger('pottery')


CacheInfo = collections.namedtuple(
    'CacheInfo',
    ('hits', 'misses', 'maxsize', 'currsize'),
)
CacheInfo.__new__.__defaults__ = 0, 0, None, 0
CacheInfo.__doc__ = ''


def _arg_hash(*args, **kwargs):
    return hash((args, frozenset(kwargs.items())))


def redis_cache(*, redis=None, key=None, timeout=_DEFAULT_TIMEOUT):
    '''Redis-backed caching decorator with an API like functools.lru_cache().

    Arguments to the original underlying function must be hashable, and return
    values from the function must be JSON serializable.

    Additionally, this decorator provides the following functions:

    f.__wrapped__(*args, **kwargs)
        Access the original underlying function.  This is useful for
        introspection, for bypassing the cache, or for rewrapping the function
        with a different cache.

    f.__bypass__(*args, **kwargs)
        Force a cache reset for your args/kwargs.  Bypass the cache lookup,
        call the original underlying function, then cache the results for
        future calls to f(*args, **kwargs).

    f.cache_info()
        Return a namedtuple showing hits, misses, maxsize, and currsize.  This
        information is helpful for measuring the effectiveness of the cache.

        Note that maxsize is always None, meaning that this cache is always
        unbounded.  maxsize is only included for compatibility with
        functools.lru_cache().

        While redis_cache() is thread-safe, also note that hits/misses only
        instrument your local process - not other processes, even if connected
        to the same Redis-backed redis_cache() key.  And in some cases,
        hits/misses may be incorrect in multiprocess/distributed applications.

        That said, currsize is always correct, even if other remote processes
        modify the same Redis-backed redis_cache() key.

    f.cache_clear()
        Clear/invalidate the entire cache (for all args/kwargs previously
        cached) for your function.

    In general, you should only use redis_cache() when you want to reuse
    previously computed values.  Accordingly, it doesn't make sense to cache
    functions with side-effects or impure functions such as time() or random().

    However, unlike functools.lru_cache(), redis_cache() reconstructs
    previously cached objects on each cache hit.  Therefore, you can use
    redis_cache() for a function that needs to create a distinct mutable object
    on each call.
    '''
    def decorator(func):
        nonlocal redis, key
        redis = Redis(socket_timeout=1) if redis is None else redis
        if key is None:  # pragma: no cover
            key = random_key(redis=redis)
            _logger.info(
                "Self-assigning key redis_cache(key='%s') for function %s",
                key,
                func.__qualname__,
            )
        cache = RedisDict(redis=redis, key=key)
        hits, misses = 0, 0

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            nonlocal hits, misses
            hash_ = _arg_hash(*args, **kwargs)
            try:
                return_value = cache[hash_]
                hits += 1
            except KeyError:
                return_value = func(*args, **kwargs)
                cache[hash_] = return_value
                misses += 1
            if timeout:
                redis.expire(key, timeout)
            return return_value

        @functools.wraps(func)
        def bypass(*args, **kwargs):
            hash_ = _arg_hash(*args, **kwargs)
            return_value = func(*args, **kwargs)
            cache[hash_] = return_value
            if timeout:
                redis.expire(key, timeout)
            return return_value

        def cache_info():
            return CacheInfo(hits=hits, misses=misses, currsize=len(cache))

        def cache_clear():
            nonlocal hits, misses
            redis.delete(key)
            hits, misses = 0, 0

        wrapper.__wrapped__ = func
        wrapper.__bypass__ = bypass
        wrapper.cache_info = cache_info
        wrapper.cache_clear = cache_clear
        return wrapper
    return decorator


class CachedOrderedDict(collections.OrderedDict):
    _SENTINEL = object()
    _NUM_TRIES = 3

    def __init__(self, *, redis=None, key=None, keys=tuple(),
                 num_tries=_NUM_TRIES):
        self._num_tries = num_tries
        partial = functools.partial(RedisDict, redis=redis, key=key)
        self._cache = self._retry(partial)
        self._misses = set()

        items = []
        for key_ in keys:
            try:
                item = (key_, self._cache[key_])
            except KeyError:
                item = (key_, self._SENTINEL)
                self._misses.add(key_)
            items.append(item)
        return super().__init__(items)

    def misses(self):
        return frozenset(self._misses)

    def __setitem__(self, key, value):
        if value is not self._SENTINEL:
            self._cache[key] = value
            self._misses.discard(key)
        return super().__setitem__(key, value)

    def setdefault(self, key, default=None):
        partial = functools.partial(
            self._retry_setdefault,
            key,
            default=default,
        )
        self._retry(partial)
        self._misses.discard(key)
        if key not in self or self[key] is self._SENTINEL:
            self[key] = default
        return self[key]

    def _retry_setdefault(self, key, default=None):
        with self._cache._watch():
            if key not in self._cache:
                self._cache.redis.multi()
                self._cache[key] = default

    def _retry(self, partial, try_num=0):
        try:
            return partial()
        except WatchError:  # pragma: no cover
            if try_num < self._num_tries - 1:
                return self._retry(partial, try_num=try_num+1)
            else:
                raise
